/* USER CODE BEGIN Header */
/**
  ******************************************************************************
  * @file           : main.c
  * @brief          : Main program body
  ******************************************************************************
  ** This notice applies to any and all portions of this file
  * that are not between comment pairs USER CODE BEGIN and
  * USER CODE END. Other portions of this file, whether 
  * inserted by the user or by software development tools
  * are owned by their respective copyright owners.
  *
  * COPYRIGHT(c) 2021 STMicroelectronics
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
  *   1. Redistributions of source code must retain the above copyright notice,
  *      this list of conditions and the following disclaimer.
  *   2. Redistributions in binary form must reproduce the above copyright notice,
  *      this list of conditions and the following disclaimer in the documentation
  *      and/or other materials provided with the distribution.
  *   3. Neither the name of STMicroelectronics nor the names of its contributors
  *      may be used to endorse or promote products derived from this software
  *      without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  ******************************************************************************
  */
/* USER CODE END Header */

/* Includes ------------------------------------------------------------------*/
#include "main.h"

/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */
#include <string.h>
#include <stdbool.h>
#include "si5351.h"
/* USER CODE END Includes */

/* Private typedef -----------------------------------------------------------*/
/* USER CODE BEGIN PTD */

/* USER CODE END PTD */

/* Private define ------------------------------------------------------------*/
/* USER CODE BEGIN PD */

/* USER CODE END PD */

/* Private macro -------------------------------------------------------------*/
/* USER CODE BEGIN PM */

/* USER CODE END PM */

/* Private variables ---------------------------------------------------------*/
I2C_HandleTypeDef hi2c1;

TIM_HandleTypeDef htim3;

/* USER CODE BEGIN PV */

/* USER CODE END PV */

/* Private function prototypes -----------------------------------------------*/
void SystemClock_Config(void);
static void MX_GPIO_Init(void);
static void MX_I2C1_Init(void);
static void MX_TIM3_Init(void);
/* USER CODE BEGIN PFP */

/* USER CODE END PFP */

/* Private user code ---------------------------------------------------------*/
/* USER CODE BEGIN 0 */

// SSB
const int32_t Fbfo = 11999900; // xtal filter highest frequency in passband + 300-500 Hz to cut off the carrier during TX
const int32_t BfoToneShift = 0; // don't shift actual BFO frequency for SSB
int32_t Fvfo = 1000000; // dummy value, will be changed below

// CW
//const int32_t Fbfo = 11997560; // xtal filter center frequency
//const int32_t BfoToneShift = -1000; // shift actual BFO frequency to get 1 KHz tone


typedef enum {
	ROTARY_ENCODER_MODE_CHANGE_FREQEUNCY = 0,
	ROTARY_ENCODER_MODE_CHANGE_STEP = 1,
} RotaryEncoderMode_t;

int32_t targetFrequency = 7115000; // SSB
int32_t frequencyStep = 1000; // SSB
// int32_t frequencyStep = 100; // CW
int32_t prevFrequencyStep = 0;
RotaryEncoderMode_t rotaryEncoderMode;
bool isTransmitting = false;

int32_t prevCounter = 0;

#define PIN_RS    (1 << 0)
#define PIN_EN    (1 << 2)
#define BACKLIGHT (1 << 3)

#define LCD_DELAY_MS 5

uint8_t lcd_addr = (0x27 << 1);

HAL_StatusTypeDef LCD_SendInternal(uint8_t data, uint8_t flags) {
    HAL_StatusTypeDef res;
    for(;;) {
        res = HAL_I2C_IsDeviceReady(&hi2c1, lcd_addr, 1,
                                    HAL_MAX_DELAY);
        if(res == HAL_OK)
            break;
    }

    uint8_t up = data & 0xF0;
    uint8_t lo = (data << 4) & 0xF0;

    uint8_t data_arr[4];
    data_arr[0] = up|flags|BACKLIGHT|PIN_EN;
    data_arr[1] = up|flags|BACKLIGHT;
    data_arr[2] = lo|flags|BACKLIGHT|PIN_EN;
    data_arr[3] = lo|flags|BACKLIGHT;

    res = HAL_I2C_Master_Transmit(&hi2c1, lcd_addr, data_arr,
                                  sizeof(data_arr), HAL_MAX_DELAY);
    HAL_Delay(LCD_DELAY_MS);
    return res;
}

void LCD_SendCommand(uint8_t cmd) {
    LCD_SendInternal(cmd, 0);
}

void LCD_SendData(uint8_t data) {
    LCD_SendInternal(data, PIN_RS);
}

void LCD_Init() {
    // 4-bit mode, 2 lines, 5x7 format
    LCD_SendCommand(0b00110000);
    // display & cursor home (keep this!)
    LCD_SendCommand(0b00000010);
    // display on, right shift, underline ON, blink off
    LCD_SendCommand(0b00001110);
    // clear display (optional here)
    LCD_SendCommand(0b00000001);
}

void LCD_SendString(char *str) {
    while(*str) {
        LCD_SendData((uint8_t)(*str));
        str++;
    }
}

int32_t frequencyStepToDigitOffset(int32_t step) {
	// returns digit offset from the left depending on frequency step
	// example:
	// 14.000.00
	//      ^
	// step = 1000
	// offset = 5
	switch(step) {
		case 1:
			return 9;
		case 10:
			return 8;
		case 100:
			return 7;
		case 1000:
			return 5;
		case 10000:
			return 4;
		case 100000:
			return 3;
		default:
			return 0;
	}
}

void changeFrequencyStep(int32_t delta) {
	int32_t step = frequencyStep;
	while(delta != 0) {
		if(delta > 0) {
			step = step / 10;
			// if(step < 100 ) {
			if(step < 1 ) {
				step = 100000;
			}
			delta--;
		} else {
			step = step * 10;
			if(step > 100000) {
				step = 1;
			}
			delta++;
		}
	}
	frequencyStep = step;
}

void changeFrequency(int32_t delta) {
	static bool pllSetupDone = false;
	si5351PLLConfig_t pll_conf;
	si5351OutputConfig_t out_conf;

	targetFrequency += frequencyStep*delta;
	if(targetFrequency < 7000000) {
		targetFrequency = 7000000;
	} else if(targetFrequency > 7200000) {
		targetFrequency = 7200000;
	}

	if(targetFrequency < 10000000) {
		Fvfo  = Fbfo-targetFrequency; // LSB
	} else {
		Fvfo = Fbfo+targetFrequency; // USB
	}
	si5351_EnableOutputs((1<<2));

	si5351_Calc(Fvfo, &pll_conf, &out_conf);
	if(!pllSetupDone) {
		// Setting up the PLL causes a brief (~1.5ms) glitch in the VFO output
		// which sounds like a loud popping sound the the speaker.
		// By setting up the PLL only once we get rid of this popping sound.
		si5351_SetupPLL(SI5351_PLL_A, &pll_conf);
		pllSetupDone = true;
	}
	si5351_SetupOutput(0, SI5351_PLL_A, SI5351_DRIVE_STRENGTH_4MA, &out_conf);

	si5351_EnableOutputs((1 << 2)|(1<<0));
}

// Swaps Fvfo and Fbfo @ CLK0 and CLK2
void enterTransmitMode() {
	si5351_EnableOutputs(0);
	si5351_SetupCLK0(Fbfo, SI5351_DRIVE_STRENGTH_4MA);
	si5351_SetupCLK2(Fvfo, SI5351_DRIVE_STRENGTH_4MA);
	si5351_EnableOutputs((1 << 2)|(1<<0));
}

void enterReceiveMode() {
	si5351_EnableOutputs(0);
	si5351_SetupCLK0(Fvfo, SI5351_DRIVE_STRENGTH_4MA);
	si5351_SetupCLK2(Fbfo, SI5351_DRIVE_STRENGTH_4MA);
	si5351_EnableOutputs((1 << 2)|(1<<0));
}

void changeRotaryEncoderMode() {
	if(rotaryEncoderMode == ROTARY_ENCODER_MODE_CHANGE_FREQEUNCY) {
		rotaryEncoderMode = ROTARY_ENCODER_MODE_CHANGE_STEP;
	} else {
		rotaryEncoderMode = ROTARY_ENCODER_MODE_CHANGE_FREQEUNCY;
	}
}

void displayCursor() {
	uint8_t digitOffset = frequencyStepToDigitOffset(frequencyStep);
    LCD_SendCommand(0b10000000 + digitOffset);
}

void displayFrequency() {
	char buff[16];
	snprintf(buff, sizeof(buff), "%02ld.%03ld.%03ld", 
		targetFrequency / 1000000, (targetFrequency % 1000000) / 1000, (targetFrequency % 1000) /* / 100 */ );

    LCD_SendCommand(0b10000000);
    LCD_SendString(buff);

	displayCursor();
}

void init() {
/*
	// Code for determining the correction factor for Si5351

	si5351_Init(0);
	si5351PLLConfig_t pll_conf;
	si5351OutputConfig_t out_conf;
	int32_t Fclk = 10000000;

	si5351_Calc(Fclk, &pll_conf, &out_conf);
	si5351_SetupPLL(SI5351_PLL_A, &pll_conf);
	si5351_SetupOutput(1, SI5351_PLL_A, SI5351_DRIVE_STRENGTH_4MA, &out_conf);
	si5351_EnableOutputs(1<<1);

	while(1) {
		HAL_Delay(100);
	}
*/

	// Check for an alternative LCD address
    HAL_StatusTypeDef res;
    res = HAL_I2C_IsDeviceReady(&hi2c1, (0x3F << 1), 1, HAL_MAX_DELAY);
    if(res == HAL_OK) {
		lcd_addr = (0x3F << 1);
	}

	HAL_TIM_Encoder_Start(&htim3, TIM_CHANNEL_ALL);

    LCD_Init();
    LCD_SendCommand(0b10000000);
    LCD_SendString(" R2AUK TRX MK 2 ");
    LCD_SendCommand(0b11000000);
    LCD_SendString("    Jan 2021    ");
	HAL_Delay(3000);
    // clear display
    LCD_SendCommand(0b00000001);

	const int32_t correction = 5477;
	si5351_Init(correction);
	si5351_SetupCLK2(Fbfo/*+BfoToneShift*/, SI5351_DRIVE_STRENGTH_4MA);
	changeFrequency(0);

    // set address to 0x00
    // LCD_SendCommand(0b10000000);
    // LCD_SendString("07.115.0 ..** S9");
    // LCD_SendString("07.115.0 SWR 1.5");

    // set address to 0x40
    LCD_SendCommand(0b11000000);
    // LCD_SendString("RX SSB  RIT+0000"); // or XIT
    LCD_SendString("RX SSB");
    
	uint8_t offset = 5;
    LCD_SendCommand(0b10000000 + offset);
    // display on, right shift, underline ON, blink off
    LCD_SendCommand(0b00001110);
	displayFrequency();
}

#define DEBOUNCE_BUTTON_A    0
#define DEBOUNCE_BUTTON_B    1
#define DEBOUNCE_BUTTON_C    2
#define DEBOUNCE_BUTTON_D    3
#define DEBOUNCE_ROT_BUTTON  4
#define DEBOUNCE_RX_TX       5

bool debounce(uint8_t buttonNumber) {
	static uint32_t lastPressed[6] = {0};
	if(buttonNumber >= (sizeof(lastPressed)/sizeof(lastPressed[0]))) {
		return false;
	}
	uint32_t now = HAL_GetTick();
	if(now - lastPressed[buttonNumber] < 200) {
		return false;
	}

	lastPressed[buttonNumber] = now;
	return true;
}

void loop() {
	int32_t currCounter = __HAL_TIM_GET_COUNTER(&htim3);
	currCounter = 32767 - ((currCounter-1) & 0xFFFF) / 2;
	if(currCounter > 32768/2) {
		// convert ... 32766, 32767, 0, 1, 2 ... into:
		//               ... -2, -1, 0, 1, 2 ...
		// this simplifies `delta` calculation
		currCounter = currCounter-32768;
	}
	if(currCounter != prevCounter) {
		int32_t delta = currCounter-prevCounter;
		prevCounter = currCounter;
		// debounce or skip counter overflow, also do nothing during TX
		if((delta > -10) && (delta < 10) && (!isTransmitting)) {
			delta = -delta; // the encoder was soldered backwards ¯\_(ツ)_/¯
			if(rotaryEncoderMode == ROTARY_ENCODER_MODE_CHANGE_FREQEUNCY) {
				changeFrequency(delta);
			} else {
				changeFrequencyStep(delta);
			}
			displayFrequency();
		}
	}

    if((HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_0) == GPIO_PIN_RESET) && debounce(DEBOUNCE_BUTTON_A)) {
		LCD_SendCommand(0b11000000 + 14);
		LCD_SendString("B1");
		displayCursor();
    }
    if((HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_1) == GPIO_PIN_RESET) && debounce(DEBOUNCE_BUTTON_B)) {
		LCD_SendCommand(0b11000000 + 14);
		LCD_SendString("B2");
		displayCursor();
    }
    if((HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_3) == GPIO_PIN_RESET) && debounce(DEBOUNCE_BUTTON_C)) {
		LCD_SendCommand(0b11000000 + 14);
		LCD_SendString("B3");
		displayCursor();
    }
    if((HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_4) == GPIO_PIN_RESET) && debounce(DEBOUNCE_BUTTON_D)) {
		LCD_SendCommand(0b11000000 + 14);
		LCD_SendString("B4");
		displayCursor();
    }

    if((HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_5) == GPIO_PIN_RESET) && debounce(DEBOUNCE_ROT_BUTTON)) {
		changeRotaryEncoderMode();
    }

	if(HAL_GPIO_ReadPin(GPIOF, GPIO_PIN_1) == GPIO_PIN_RESET) {
		if((!isTransmitting) && debounce(DEBOUNCE_RX_TX)) {
			LCD_SendCommand(0b11000000);
			LCD_SendString("T");
			displayCursor();

			enterTransmitMode();
			isTransmitting = true;
		}
	} else {
		if(isTransmitting && debounce(DEBOUNCE_RX_TX)) {
			LCD_SendCommand(0b11000000);
			LCD_SendString("R");
			displayCursor();

			enterReceiveMode();
			isTransmitting = false;
		}
	}

    HAL_Delay(100);
}
/* USER CODE END 0 */

/**
  * @brief  The application entry point.
  * @retval int
  */
int main(void)
{
  /* USER CODE BEGIN 1 */

  /* USER CODE END 1 */

  /* MCU Configuration--------------------------------------------------------*/

  /* Reset of all peripherals, Initializes the Flash interface and the Systick. */
  HAL_Init();

  /* USER CODE BEGIN Init */

  /* USER CODE END Init */

  /* Configure the system clock */
  SystemClock_Config();

  /* USER CODE BEGIN SysInit */

  /* USER CODE END SysInit */

  /* Initialize all configured peripherals */
  MX_GPIO_Init();
  MX_I2C1_Init();
  MX_TIM3_Init();
  /* USER CODE BEGIN 2 */

  init();
  /* USER CODE END 2 */

  /* Infinite loop */
  /* USER CODE BEGIN WHILE */
  while (1)
  {
    loop();
    /* USER CODE END WHILE */

    /* USER CODE BEGIN 3 */
  }
  /* USER CODE END 3 */
}

/**
  * @brief System Clock Configuration
  * @retval None
  */
void SystemClock_Config(void)
{
  RCC_OscInitTypeDef RCC_OscInitStruct = {0};
  RCC_ClkInitTypeDef RCC_ClkInitStruct = {0};
  RCC_PeriphCLKInitTypeDef PeriphClkInit = {0};

  /**Initializes the CPU, AHB and APB busses clocks 
  */
  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSI;
  RCC_OscInitStruct.HSIState = RCC_HSI_ON;
  RCC_OscInitStruct.HSICalibrationValue = RCC_HSICALIBRATION_DEFAULT;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_NONE;
  if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
  {
    Error_Handler();
  }
  /**Initializes the CPU, AHB and APB busses clocks 
  */
  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK|RCC_CLOCKTYPE_SYSCLK
                              |RCC_CLOCKTYPE_PCLK1;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_HSI;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV1;

  if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_0) != HAL_OK)
  {
    Error_Handler();
  }
  PeriphClkInit.PeriphClockSelection = RCC_PERIPHCLK_I2C1;
  PeriphClkInit.I2c1ClockSelection = RCC_I2C1CLKSOURCE_HSI;
  if (HAL_RCCEx_PeriphCLKConfig(&PeriphClkInit) != HAL_OK)
  {
    Error_Handler();
  }
}

/**
  * @brief I2C1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_I2C1_Init(void)
{

  /* USER CODE BEGIN I2C1_Init 0 */

  /* USER CODE END I2C1_Init 0 */

  /* USER CODE BEGIN I2C1_Init 1 */

  /* USER CODE END I2C1_Init 1 */
  hi2c1.Instance = I2C1;
  hi2c1.Init.Timing = 0x2000090E;
  hi2c1.Init.OwnAddress1 = 0;
  hi2c1.Init.AddressingMode = I2C_ADDRESSINGMODE_7BIT;
  hi2c1.Init.DualAddressMode = I2C_DUALADDRESS_DISABLE;
  hi2c1.Init.OwnAddress2 = 0;
  hi2c1.Init.OwnAddress2Masks = I2C_OA2_NOMASK;
  hi2c1.Init.GeneralCallMode = I2C_GENERALCALL_DISABLE;
  hi2c1.Init.NoStretchMode = I2C_NOSTRETCH_DISABLE;
  if (HAL_I2C_Init(&hi2c1) != HAL_OK)
  {
    Error_Handler();
  }
  /**Configure Analogue filter 
  */
  if (HAL_I2CEx_ConfigAnalogFilter(&hi2c1, I2C_ANALOGFILTER_ENABLE) != HAL_OK)
  {
    Error_Handler();
  }
  /**Configure Digital filter 
  */
  if (HAL_I2CEx_ConfigDigitalFilter(&hi2c1, 0) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN I2C1_Init 2 */

  /* USER CODE END I2C1_Init 2 */

}

/**
  * @brief TIM3 Initialization Function
  * @param None
  * @retval None
  */
static void MX_TIM3_Init(void)
{

  /* USER CODE BEGIN TIM3_Init 0 */

  /* USER CODE END TIM3_Init 0 */

  TIM_Encoder_InitTypeDef sConfig = {0};
  TIM_MasterConfigTypeDef sMasterConfig = {0};

  /* USER CODE BEGIN TIM3_Init 1 */

  /* USER CODE END TIM3_Init 1 */
  htim3.Instance = TIM3;
  htim3.Init.Prescaler = 0;
  htim3.Init.CounterMode = TIM_COUNTERMODE_UP;
  htim3.Init.Period = 65535;
  htim3.Init.ClockDivision = TIM_CLOCKDIVISION_DIV1;
  htim3.Init.AutoReloadPreload = TIM_AUTORELOAD_PRELOAD_DISABLE;
  sConfig.EncoderMode = TIM_ENCODERMODE_TI1;
  sConfig.IC1Polarity = TIM_ICPOLARITY_RISING;
  sConfig.IC1Selection = TIM_ICSELECTION_DIRECTTI;
  sConfig.IC1Prescaler = TIM_ICPSC_DIV1;
  sConfig.IC1Filter = 0;
  sConfig.IC2Polarity = TIM_ICPOLARITY_RISING;
  sConfig.IC2Selection = TIM_ICSELECTION_DIRECTTI;
  sConfig.IC2Prescaler = TIM_ICPSC_DIV1;
  sConfig.IC2Filter = 0;
  if (HAL_TIM_Encoder_Init(&htim3, &sConfig) != HAL_OK)
  {
    Error_Handler();
  }
  sMasterConfig.MasterOutputTrigger = TIM_TRGO_RESET;
  sMasterConfig.MasterSlaveMode = TIM_MASTERSLAVEMODE_DISABLE;
  if (HAL_TIMEx_MasterConfigSynchronization(&htim3, &sMasterConfig) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN TIM3_Init 2 */

  /* USER CODE END TIM3_Init 2 */

}

/**
  * @brief GPIO Initialization Function
  * @param None
  * @retval None
  */
static void MX_GPIO_Init(void)
{
  GPIO_InitTypeDef GPIO_InitStruct = {0};

  /* GPIO Ports Clock Enable */
  __HAL_RCC_GPIOF_CLK_ENABLE();
  __HAL_RCC_GPIOA_CLK_ENABLE();

  /*Configure GPIO pin : PF1 */
  GPIO_InitStruct.Pin = GPIO_PIN_1;
  GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
  GPIO_InitStruct.Pull = GPIO_PULLUP;
  HAL_GPIO_Init(GPIOF, &GPIO_InitStruct);

  /*Configure GPIO pins : PA0 PA1 PA3 PA4 
                           PA5 */
  GPIO_InitStruct.Pin = GPIO_PIN_0|GPIO_PIN_1|GPIO_PIN_3|GPIO_PIN_4 
                          |GPIO_PIN_5;
  GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
  GPIO_InitStruct.Pull = GPIO_PULLUP;
  HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

}

/* USER CODE BEGIN 4 */

/* USER CODE END 4 */

/**
  * @brief  This function is executed in case of error occurrence.
  * @retval None
  */
void Error_Handler(void)
{
  /* USER CODE BEGIN Error_Handler_Debug */
  /* User can add his own implementation to report the HAL error return state */

  /* USER CODE END Error_Handler_Debug */
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(char *file, uint32_t line)
{ 
  /* USER CODE BEGIN 6 */
  /* User can add his own implementation to report the file name and line number,
     tex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */
  /* USER CODE END 6 */
}
#endif /* USE_FULL_ASSERT */

/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/
